// rgtp.js -- implements the client side of RGTP
// $Header: /cvs/archangel/src/archangel/content/rgtp.js,v 1.9 2003/02/11 07:38:56 marnanel Exp $
//
// Copyright (c) 2003 Thomas Thurman
// thomas@thurman.org.uk
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have be able to view the GNU General Public License at 
// http://www.gnu.org/copyleft/gpl.html ; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
//
////////////////////////////////////////////////////////////////
//
// This is the RGTP library. It provides a consistent and simple interface
// to the RGTP protocol.
//
// All the client-callable methods are asynchronous: you call them and
// they return instantly. The job will complete in the background.
// In case you want to know when the work is compete, you may call
// evalAfterNextJob() to provide a "clearup" string, which will be
// eval()ed once the job is done.
//
// The code falls into three parts:
//  1. Public methods for the library: if you're a client,
//     you just need to know this bit. The methods include:
//
//           open()     - open connection to RGTP server
//           getItem()  - retrieve RGTP item
//           getIndex() - retrieve RGTP index
//
//  2. "Line handler" methods.
//  3. Code to send and receive from the TCP connection.
//
// "Line handlers" are routines which process incoming lines from the
// RGTP server; they all have names beginning with "handle".
// Only one is active at any time, and that one can be called using
// the method handleLine(), which always points at the correct one.
//
// Each handler takes two parameters. The first is the RGTP code
// (see the RGTP protocol spec for details) and the second is the
// message portion of the line. There are two special cases of the
// first parameter: if it is -1, the server is sending us a line
// from a file. If it is 0, the server has just finished sending
// the file.
//
// The private method doneTheJob() should be called by methods when
// they have finished their job. It resets handleLine(), and execs
// and clears the clearup string if necessary.
//
// It would be rather amazingly useful to split off the code which
// handles line-by-line socket handling into a separate file.

////////////////////////////////////////////////////////////////
//
//                         ANCILLARY CODE
//
////////////////////////////////////////////////////////////////

// Returns the number |number| formatted in hex with leading zeroes
// up to a width of 8. (In printf-language, this is "%08x".)
function eighthex(number) {
	var hex =  number.toString(16);
	while (hex.length<8) hex = '0'+hex;
	return hex;
}

function trim(str) { return str.replace(/\s*$/,''); }

//////////////////////////////////////////////////////////////////
//
//                         PUBLIC METHODS
//
////////////////////////////////////////////////////////////////

RGTPConnection.prototype.evalAfterNextJob =
function rc_evalAfterNextJob(clearupString) {
	this.afterThisJob = clearupString;
}

var main_rgtp = 0; // icky, but it'll have to do for now

RGTPConnection.prototype.open =
function rc_open() {
	this.handleLine = this.handleInitialConnection;
	this.handlingFile = 0;

	this.server = new CConnection(this.host, this.port);
	this.server.onRead = function(b) { main_rgtp.processBlock(b); }
	this.server.onClose = function(c,m) { main_rgtp.reportClosedConnection(c,m); }
	var result = this.server.open();
	this.valid = 1;
}

RGTPConnection.prototype.getItem =
function rc_getItem(itemid) {

	stripyMeter('Downloading item '+itemid);

	this.send('ITEM '+itemid);
	this.handleLine = this.handleItem;
}

RGTPConnection.prototype.getIndex =
function rc_getIndex() {

	// If we've seen some things already,
	// then we'll have stored "maxknown", the highest
	// sequence number we know.

	maxknown = this.node.getAttribute('maxknown');
	if (maxknown) {
		stripyMeter('Downloading partial index');
		// No point warning them about long
		// download times; this should only take a
		// few seconds, if that.

		this.send('INDX #'+maxknown);
	} else {
		stripyMeter('Downloading full index');
		alert('Archangel must now download the RGTP index. ' +
			'The download may take a few minutes, since '+
			'the index is typically half a megabyte long.');
		this.send('INDX');
	}

	this.handleLine = this.handleIndex;
}

//////////////////////////////////////////////////////////////////
//
//                         LINE HANDLERS
//
////////////////////////////////////////////////////////////////

RGTPConnection.prototype.handleInitialConnection =
function rc_handleInitialConnection(code, text) {

	// does a rather bad job of general RGTP initial connection
	// handling, but it's enough for GROGGS.

	if (code=='230') {
		this.access = 0;
		this.send('USER guest'); // for now, anyway
	} else if (code=='231') {
		// Okay, so we're in.
		this.access = 1;
		this.doneTheJob();
	} else
		throw "weird code in lobby";
}

RGTPConnection.prototype.handleUnexpected =
function rc_handleUnexpected(code, text) {
	alert('Unexpected message received from RGTP server:\n\n'+
		code+' '+text+'\n\nIgnoring.');
}

var mostRecentlySeenChild = 0;

RGTPConnection.prototype.handleIndex =
function rc_handleIndex(code, text) {

	if (code==0) {
		var typecode = text.charAt(103);
		var itemid = text.substring(18,26);
		var author = trim(text.substring(27,101));
		var datestamp = eval('0x'+text.substring(9,17));
		var sequence = eval('0x'+text.substring(0,8));

		this.node.setAttribute('maxknown', eighthex(sequence+1));

		if (mostRecentlySeenChild) {

			// Then we've just seen a C, so this is an F.

			if (typecode!='F')
				throw 'F must follow C in the index, but #'+
					eighthex(sequence)+' is '+typecode+'!';

			var parent = this.node.getNode(itemid);
			var child = mostRecentlySeenChild;

			parent.setAttribute('child', child.getSubject());
			child.setAttribute('parent', parent.getSubject());
				
			mostRecentlySeenChild = 0;

		} else if (typecode=='R') {
			var itemnode = this.node.getNode(itemid);

			itemnode.setAttribute('known', eighthex(sequence));
			itemnode.setAttribute('last-by', author);
			itemnode.setAttribute('last-date', datestamp);
			itemnode.setAttribute('replies',
				itemnode.getAttribute('replies')*1+1);
		} else if (typecode=='I' || typecode=='C') {
			var subject = trim(text.substring(105));
			var itemnode = this.node.addNode(itemid);

			itemnode.setAttribute('subject', subject);
			itemnode.setAttribute('first-by', author);
			itemnode.setAttribute('last-by', author);
			itemnode.setAttribute('first-date', datestamp);
			itemnode.setAttribute('last-date', datestamp);
			itemnode.setAttribute('known', eighthex(sequence));
			itemnode.setAttribute('replies', 0);

			// For "C", we also need to store this item node so
			// that we can set the parent-child relationship
			// correctly. The next record must be an "F",
			// representing the parent item of this one.

			if (typecode=='C')
				mostRecentlySeenChild = itemnode;
		} else if (typecode=='M') {
			var motdNode = this.node.getNode('motd');
			if (motdNode==null)
				motdNode = this.node.addNode('motd');

			motdNode.setAttribute('known', eighthex(sequence));
		} else if (typecode=='E') {
			// just ignore these
		} else {
			throw 'Unknown typecode '+typecode+' at #'+
				eighthex(sequence);
		}

	} else if (code==-1) {
		this.node.flush();
		this.doneTheJob();
	} else if (code!=250)
		throw "weird code while reading index";
}

RGTPConnection.prototype.handleItem =
function rc_handleItem(code, text) {
	if (code==0) {
		receivedItemLine(text);
	} else if (code==-1) {
		// it's all over
		this.node.flush();
		this.doneTheJob();
	} else if (code!=250)
		throw "weird code while reading item";
}

//////////////////////////////////////////////////////////////////
//
//                         TCP INTERFACE
//
////////////////////////////////////////////////////////////////

RGTPConnection.prototype.send =
function rc_send(text) {
	this.server.write(text+'\r\n');
}

RGTPConnection.prototype.processBlock =
function rc_processBlock(block) {
	try {
		if (this.partialLines) {
			this.partialLines = this.partialLines + block;
		} else {
			this.partialLines = block;
		}

		n = 0;

		while (n!=-1) {
			n = this.partialLines.indexOf('\r\n');
			if (n!=-1) {
				var current = this.partialLines.substring(0, n);

				if (this.handlingFile) {
					if (current=='.') {
						this.handlingFile = 0;
						this.handleLine(-1, current);
					} else {
						this.handleLine(0, current);
					}
				} else {
					this.handleLine(
						current.substring(0,3),
						current.substring(4));
	
					if (current.substring(0,3) == '250')
						this.handlingFile = 1;
				}
				this.partialLines = this.partialLines.substring(n+2);
			}
		}
	} catch (e) {
		alert('RGTP internal error: '+e);
		throw e;
	}

}

//////////////////////////////////////////////////////////////////

RGTPConnection.prototype.doneTheJob =
function rc_doneTheJob() {
	calmMeter();
	this.handleLine = this.handleUnexpected;
	if (this.afterThisJob) {
		var temp = this.afterThisJob;

		this.afterThisJob = '';
		eval(temp);
	}
}


RGTPConnection.prototype.reportClosedConnection =
function rc_reportClosedConnection(code, message) {

	this.valid = 0;

	var say = '';
	
	switch(code) {
		case NS_ERROR_UNKNOWN_HOST:
			say = "Couldn't find RGTP server "+this.host+".";
			break;

		case NS_ERROR_CONNECTION_REFUSED:
			say = "The server "+this.host+' is not accepting ';
			if (main_rgtp.port = 1491)
				say = say + 'RGTP connections.';
			else
				say = say + 'connections on port '+this.port;
			break;

		case NS_ERROR_NET_TIMEOUT:
			say = "The server "+this.host+" did not respond in time.";
			break;
			
		default:
			say = 'Unexpected network error: '+message;	
	}

	alert(say);
	this.doneTheJob(); // well, pretty much anyway
}


////////////////////////////////////////////////////////////////

// Returns a connection to an RGTP server.

function RGTPConnection(rdf_node) {

	this.node = rdf_node;
	this.host = rdf_node.getAttribute('host');
	this.port = rdf_node.getAttribute('port');
	this.valid = 0;
	this.afterThisJob = '';
	this.partialLines = '';

	main_rgtp = this;

	return this;
}



